// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "UILabelNew.hpp"
#include "game/DawnGame.hpp"

using namespace Dawn;

UILabelNew::UILabelNew(SceneItem *item) :
  UIComponentRenderable(item)
{

}

void UILabelNew::rebufferQuads(std::vector<struct UILabelText> texts) {
  std::cout << "Rebuffering" << std::endl;
  auto oldTexts = this->texts;

  textureMap.clear();
  glm::vec2 position(0, 0);
  struct FontShaderBufferData fontData;
  int32_t quadIndex = 0;
  int32_t partIndex = 0;
  int32_t quadCount = 0;
  int32_t nextTexture = 0;

  // Determine how many quads there are, and the texture indexes.
  auto itText = texts.begin();
  while(itText != texts.end()) {
    quadCount += itText->text.length();

    // Determine font and lock it.
    assertNotNull(itText->font);
    itText->lockId = itText->font->lock(NewTrueTypeFaceTextureStyle{
      itText->size,
      itText->style
    });
    assertTrue(itText->lockId != -1);
    itText->texture = itText->font->getTexture(itText->lockId);

    // Check for existing texture, if not, map it.
    if(textureMap.find(itText->texture) == textureMap.end()) {
      assertTrue(nextTexture < FONT_SHADER_TEXTURE_MAX);
      textureMap[itText->texture] = nextTexture++;
    }
  
    // Set initial line height
    position.y = mathMax<float_t>(itText->size, position.y);

    ++itText;
  }

  // Cleanup old texst, we do this second so we don't unlock, cleanup, and then
  // lock the same font, causing it to have to re-load.
  itText = oldTexts.begin();
  while(itText != oldTexts.end()) {
    assertTrue(itText->lockId != -1);
    assertNotNull(itText->font);
    itText->font->unlock(itText->lockId);
    ++itText;
  }

  // Update texts.
  this->texts = texts;
  
  // Create mesh
  this->mesh.createBuffers(
    QUAD_VERTICE_COUNT * quadCount,
    QUAD_INDICE_COUNT * quadCount
  );

  // Buffer the text quads
  itText = texts.begin();
  while(itText != texts.end()) {
    quadIndex += this->bufferQuads(
      *itText,
      fontData,
      textureMap,
      position,
      quadIndex,
      partIndex
    );
    partIndex++;
    ++itText;
  }

  shaderBuffer.buffer(&fontData);
}

void UILabelNew::onStart() {
  this->shaderBuffer.init();

  auto font = this->getGame()->assetManager.get<NewTrueTypeAsset>("font_arial");

  std::vector<struct UILabelText> texts;
  texts.push_back({
    .text = "Hello",
    .color = COLOR_RED,
    .style = 0,
    .size = 16,
    .font = font
  });
  texts.push_back({
    .text = " World",
    .color = COLOR_BLUE,
    .style = 1,
    .size = 32,
    .font = font
  });
  this->rebufferQuads(texts);

  // this->texts.clear();

  // useEffect([&]{
  //   usagelockid_t newLock = -1;

  //   if(font != nullptr) {
  //     newLock = font->lock(NewTrueTypeFaceTextureStyle{
  //       fontSize,
  //       style
  //     });
  //   }

  //   if(fontLock != -1) {
  //     font.previous->unlock(fontLock);
  //     fontLock = -1;
  //   }

  //   if(font == nullptr) return;

  //   fontLock = newLock;
  //   this->rebufferQuads();
  // }, font)();

  // useEffect([&]{
  //   if(font == nullptr) return;
  //   auto newLock = font->lock(NewTrueTypeFaceTextureStyle{
  //     fontSize,
  //     style
  //   });

  //   if(fontLock != -1) {
  //     font->unlock(fontLock);
  //     fontLock = -1;
  //   }

  //   fontLock = newLock;
  //   this->rebufferQuads();
  // }, { &fontSize, &style })();
}

std::vector<struct ShaderPassItem> UILabelNew::getUIRenderPasses() {
  if(this->texts.size() == 0) return {};

  auto canvas = this->getCanvas();
  auto shader = getGame()->renderManager.fontShader;

  struct ShaderPassItem item;
  item.shader = shader;
  item.mesh = &this->mesh;
  item.matrixValues[shader->paramModel] = transform->getWorldTransform();
  item.parameterBuffers[shader->bufferUiCanvas] = &canvas->shaderBuffer;
  item.parameterBuffers[shader->bufferFont] = &this->shaderBuffer;
  item.renderFlags = RENDER_MANAGER_RENDER_FLAG_BLEND;

  // Map texture slots
  auto it = textureMap.begin();
  while(it != textureMap.end()) {
    item.textureSlots[it->second] = &it->first->texture;
    
    shaderparameter_t param = FONT_SHADER_TEXTURE_MAX;
    switch(it->second) {
      case 0:
        param = shader->paramTexture0;
        break;
      
      case 1:
        param = shader->paramTexture1;
        break;

      case 2:
        param = shader->paramTexture2;
        break;

      case 3:
        param = shader->paramTexture3;
        break;

      default:
        assertUnreachable();
    }
    item.textureValues[param] = it->second;
    ++it;
  }

  return { item };
}

float_t UILabelNew::getWidth() {
  return 0;
}

float_t UILabelNew::getHeight() {
  return 0;
}

float_t UILabelNew::getContentWidth() {
  return 0;
}

float_t UILabelNew::getContentHeight() {
  return 0;
}

int32_t UILabelNew::bufferQuads(
  struct UILabelText text,
  struct FontShaderBufferData &bufferData,
  std::map<NewTrueTypeFaceTexture*, int32_t> &textureMap,
  glm::vec2 &position,
  int32_t quadStart,
  int32_t partIndex
) {
  // Get string length
  int32_t len = text.text.length();
  
  glm::vec2 wh = glm::vec2(
    text.texture->texture.getWidth(),
    text.texture->texture.getHeight()
  );

  // For each char
  for(int32_t i = 0; i < len; i++) {
    char ch = text.text[i];
    int32_t j = quadStart + i;
    FT_ULong c = ch;
    auto charInfo = text.texture->getCharacterData(c);

    // Determine texture coordinates.
    glm::vec2 uv0 = glm::vec2(0.0f, charInfo.textureY) / wh;
    glm::vec2 uv1 = uv0 + (charInfo.bitmapSize / wh);

    // Buffer the quad.
    QuadMesh::bufferQuadMeshWithZ(&this->mesh,
      position + charInfo.bitmapPosition, uv0,
      position + charInfo.bitmapPosition + charInfo.bitmapSize, uv1,
      0.0f,
      j * QUAD_VERTICE_COUNT, j * QUAD_INDICE_COUNT
    );

    // Move the current position along.
    position.x += charInfo.advanceX;
    position.y += charInfo.advanceY;

    // Set the part index to the quad mappings
    bufferData.fontQuadMappings[j] = partIndex;
  }

  // Map texture level values
  bufferData.colors[partIndex] = text.color;

  auto textureId = textureMap.find(text.texture);
  assertTrue(textureId != textureMap.end());
  bufferData.textures[partIndex] = textureId->second;

  return len;
}